# This app should check for the presence of certain
# cookies, and optionally validate their values.
#
# It should either show a login page, or the requested path
require 'logger'

module Skypager
  class Protector
    attr_accessor :validator, :options, :app, :login_handler

    # Skypager::Protector.new(options) do |request|
    #   # validate request here
    # end
    #
    # Valid options are:
    #   - build_dir: defaults to ./build
    #   - login_path: defaults to /login/index.html
    #   - not_found_path: defaults to /not-found.html
    def initialize(app, options={}, &validator)
      @app = app
      @options = options
      @validator = validator
      @logger = options.fetch(:logger) { Logger.new(STDOUT) }
    end

    def log message
      @logger && @logger.info(message)
      puts message if !@logger
    end

    def call(env)
      request = Rack::Request.new(env)

      if request.post?
        log "#{ login_path } #{ request.path } #{ login_handler.respond_to?(:call) }"
      end

      if request.post? && request.path == login_path && login_handler.respond_to?(:call)

        result = login_handler.call(request)
        return serve_login_path if result == false

        redirect_path = request.params['_redirect_to'] || "/"

        return redirect_to(redirect_path, result[:cookies] || {})
      end

      if !validate(request) && request.path != login_path
        log "== Validation failed: #{ request.path }"
        return serve_login_path
      end

      path = normalize_path(request.path)
      file = build_path.join(path)

      if file.exist?
        serve_file(file)
      else
        @app.call(env)
      end
    end

    def serve_file(file, status=200)
      mime_type = Rack::Mime.mime_type(Pathname(file).extname).to_s
      content   = Pathname(file).read.to_s
      content_length = Rack::Utils.bytesize(content).to_s

      [status, {"Content-Type" => mime_type,"Content-Length" => content_length}, [content]]
    end

    def build_path
      build_dir =  options.fetch(:build_dir) { Pathname(Dir.pwd).join('build') }
      Pathname(build_dir)
    end

    def serve_not_found_path
      serve_file(not_found_path, 404)
    end

    def redirect_to(path, cookies={})
      status, headers, body = [302, {"Content-Type" => "text/html", "Location" =>  path}, ["302 You've been redirected"]]
      response = Rack::Response.new(body,status,headers)

      unless cookies.empty?
        cookies.each do |k,v|
          response.set_cookie(k, cookie_options.merge(:value=>v))
        end
      end

      response.finish
    end

    def cookie_options
      options.fetch(:cookie_options) do
        {
          path: "/",
          expires: 180.minutes.from_now,
          domain: ".lvh.me"
        }
      end
    end

    def serve_login_path
      redirect_to(login_path)
    end

    def not_found_path
       options.fetch(:not_found_path, "/not-found.html")
    end

    def login_path
      options.fetch(:login_path, "/login.html")
    end

    def normalize_path(path)
      if path == "/"
        path = "/index.html"
      elsif !path.to_s.match(/\.\w+/)
        path.gsub!(/\/$/,'')
        path + path_extension
      end

      path.to_s.gsub(/^\//,'')
    end

    def path_extension
      options.fetch(:path_extension ) do
        options[:directory_indexes] ? "/index.html" : ".html"
      end
    end

    def validate(request)
      return true if validator.nil? || !validator.respond_to?(:call)

      return true if whitelist && request.path.match(whitelist)

      validator.call({
        cookies: request.cookies,
        params: request.params,
        env: request.env
      })
    end

    def whitelist
      options.fetch(:whitelist) { /^\/(js|css|img|fonts|favico)/ }
    end
  end

end
