# Based on the s3_proxy gem
require 'rack/proxy'
module Skypager
  class Proxy

    attr_reader :bucket,
                :options,
                :access_key_id,
                :secret_access_key

    def initialize(options={})
      @options            = options
      @bucket             = options[:bucket]
    end

    def call(env)
      if @options[:bucket]
        s3_call(env)
      elsif @options[:host] || @options[:port]
        proxy_call(env)
      end
    end

    def proxy_call(env)
      request_proxy.call(env)
    end

    def s3
      Datapimp::Sync.amazon.storage
    end

    def request_proxy
      o = @options
      @request_proxy ||= RequestProxy.new.tap {|p| p.set_config(o) }
    end

    def s3_call(env)
      return Errors.method_not_allowed unless %w(GET HEAD).include?(env['REQUEST_METHOD'])
      return Errors.not_found('path_info') if env['PATH_INFO'].empty?

      rack_request = Rack::Request.new(env)
      key = "#{rack_request.path}".without_leading_slash

      return Errors.not_found('no-bucket-no-key') unless bucket && key

      path = key.without_leading_slash
      last = path.split("/").last

      req = {bucket: bucket, key: path}

      req[:key]                   = "index.html" if req[:key].to_s.length == 0

      req['If-Match']              = env['HTTP_IF_MATCH'] if env['HTTP_IF_MATCH']
      req['If-None-Match']         = env['HTTP_IF_NONE_MATCH'] if env['HTTP_IF_NONE_MATCH']
      req['If-Modified-Since']     = env['HTTP_IF_MODIFIED_SINCE'] if env['HTTP_IF_MODIFIED_SINCE']
      req['If-Unmodified-Since']   = env['HTTP_IF_UNMODIFIED_SINCE'] if env['HTTP_UNMODIFIED_SINCE']

      begin
        head = s3.head_object(bucket, key, req)
      rescue => e
        return Errors.not_found('could not head ' + bucket.to_s + ' ' + key.to_s + ' ' + e.message)
      end

      return Errors.not_found('no-head') unless head

      case env['REQUEST_METHOD']
      when 'HEAD'
        gentle env, req, head
      when 'GET'
        if env['rack.hijack?']
          hijack env, req, head
        else
          gentle env, req, head
        end
      end

    rescue Excon::Errors::NotFound
      return Errors.not_found('rescue')

    rescue Excon::Errors::NotModified
      return Errors.not_modified

    rescue Excon::Errors::PreconditionFailed
      return Errors.precondition_failed

    rescue NameError => e
      # https://github.com/aws/aws-sdk-core-ruby/pull/65
      raise e unless e.message == "wrong constant name 412Error"

      return Errors.precondition_failed
    end

    private

    def hijack(env, request, head)
      env['rack.hijack'].call

      io = env['rack.hijack_io']
      begin
        io.write "HTTP/1.1 200 OK\r\n"
        io.write "Status: 200\r\n"
        io.write "Connection: close\r\n"
        io.write "Content-Type: #{head.headers['Content-Type']}\r\n"
        io.write "Content-Length: #{head.headers['Content-Length']}\r\n"
        io.write "ETag: #{head.headers['ETag']}\r\n"
        io.write "Last-Modified: #{head.headers['Last-Modified']}\r\n"
        io.write "\r\n"
        io.flush

        s3.get_object(request, target: io)
      ensure
        io.close
      end
      return [200, {}, ['']]
    end

    def gentle(env, request, head)
      case env['REQUEST_METHOD']
      when 'GET'
        fiber = Fiber.new do
          s3.get_object(request[:bucket], request[:key], request) do |chunk|
            Fiber.yield(chunk)
          end
          Fiber.yield(nil)
        end

        body = Enumerator.new do |y|
          while n = fiber.resume
            y << n
          end
        end
      when 'HEAD'
        body = ['']
      end

      headers = {
        'Content-Type' => head.headers['Content-Type'].to_s,
        'Content-Length' => head.headers['Content-Length'].to_s,
        'Last-Modified' => head.headers['Last-Modified'].to_s,
        'ETag' => head.headers['ETag'].to_s,
      }

      [200, headers, body]
    end

    module Errors
      class << self
        def method_not_allowed
          [405, {'Content-Type' => 'text/plain'}, ["method not allowed"]]
        end

        def not_found(reason="not found man")
          [404, {'Content-Type' => 'text/plain'}, [reason.to_s]]
        end

        def forbidden
          [403, {'Content-Type' => 'text/plain'}, ["forbidden"]]
        end

        def precondition_failed
          [412, {'Content-Type' => 'text/plain'}, ["precondition failed"]]
        end

        def not_modified
          [304, {}, [""]]
        end

        def unknown(code)
          [code, {'Content-Type' => 'text/plain'}, ["Error: #{code}"]]
        end
      end
    end

    class RequestProxy < Rack::Proxy
      def set_config options={}
        cfg.merge!(options)
        cfg
      end

      def cfg
        @cfg ||= {}
      end

      def rewrite_env(e)
        e.tap do |env|
          current = env['HTTP_HOST']

          current_host, current_port = current.split(':')

          host = cfg[:host].to_s
          host = host + ":#{ cfg[:port] }" if cfg[:port]

          Rails.logger.info "replacing #{ current } with #{ host } #{ env['PATH_INFO'] }"
          env['HTTP_HOST'] = host
        end
      end
    end
  end
end
