module Skypager
  class Site
    class << self
      def directory= value
        value = value.to_mash if value.respond_to?(:to_mash)
        @directory = Skypager.config.home_config['sites_directory'] = {}.to_mash
      end

      def directory
        unless Skypager.config.home_config.key?('sites_directory')
          self.directory = {}
        end

        @directory ||= begin
                        val = Skypager.config.home_config['sites_directory'] ||= {}.to_mash
                        val = val.to_mash if val.is_a?(Hash)
                        val
                       end
      end

      def sites
        directory.values
      end

      def requiring_build
        sites.select {|site| site.requires_build == "yes" }
      end

      def find_sites_for_dropbox_uid uid
        configs = sites.select do |site|
          syncables = site.syncables
          dropbox   = syncables.dropbox ||= {}.to_mash

          uid.to_s.length > 0 && dropbox.uid.to_s == uid.to_s
        end

        configs.map {|cfg| Skypager::Site.new(cfg.name, config: cfg) }
      end

      def find_site_by_dropbox uid, path_prefix
        find_sites_for_dropbox_uid(uid).find do |site|
          Array(site.syncables.dropbox.paths).any? do |path|
            path == path_prefix || path.match(/^#{ path_prefix }/)
          end
        end
      end
    end

    attr_accessor :options,
                  :name,
                  :config


    def initialize(name, options = {})
      @name     = name.to_s.downcase
      @options  = options
      @config   = options.delete(:config)
      @config   ||= self.class.directory[@name] ||= {name: @name}
      @config   = @config.to_mash if @config.is_a?(Hash)

      @config.reverse_merge!(options[:deploy_options]) if options[:deploy_options]
    end

    def build_command
      "bundle exec middleman build"
    end

    def syncables
      config[:syncables].try(:to_mash)
    end

    def delete!
      directory = self.class.directory
      directory.delete(name)
      save!
    end

    def save!
      directory = Skypager.config.home_config["sites_directory"] || {}
      directory[name] = to_config
      Skypager.config.save_home_config
    end

    def root
      @config[:root].to_s.length > 0 && Pathname(@config[:root])
    end

    def load_config
      if existing = self.class.directory[name]
        @config = existing.to_mash
      end
    end

    def config
      @config = @config.to_mash if @config.is_a?(Hash)
      @config ||= {}.to_mash
    end

    def to_config
      config
    end

    def require_build!(state=true)
      requires_build!(state)
    end

    def requires_build!(state=true)
      state = !!(state) ? "yes" : "no"
      set(:requires_build, state)
    end

    def requires_build?
      value = get!(:requires_build)
      value == "yes"
    end

    def set attribute, value
      @config[attribute.to_sym] = value
      save!
      self
    end

    def get! attribute
      val = get(attribute)

      if val.nil? || val.to_s.length == 0
        nil
      else
        val
      end
    end

    def get attribute
      @config[attribute.to_sym]
    end

    def deploy_options
      options.fetch(:deploy_options) do
        (config.deploy_options || {}).to_mash
      end
    end

    def hostname= value
      set(:hostname, value)
    end

    def hostname
      custom_domain || bucket_hostname
    end

    def bucket_hostname
      "#{@config[:name] || @name}.#{ Skypager.config.domain }"
    end

    def bucket(create=false)
      return @bucket if @bucket

      val = deploy_options.fetch(:bucket_name) { bucket_hostname }

      b = Skypager.amazon.find_bucket_by_name(val) rescue nil

      if create && !b
        b = Skypager.amazon.create_bucket(val)
      end

      @bucket = b
    end

    def prepare_deploy(build_dir)

      @deploy_targets = []

      upload = lambda do |node|
        next if node.to_s == ".DS_Store"

        if node.directory?
          Array(node.children).each(&upload)
        elsif node.file?
          rel = node.relative_path_from(build_dir).to_s
          previous = build_manifest[rel]
          current = Digest::MD5.hexdigest(node.read)

          if current != previous
            @deploy_targets << {key: rel, node: node, digest: current}
          end
        end
      end

      build_dir.children.each(&upload)

      @deploy_targets
    end

    def deploy(build_dir)
      target = bucket || Skypager.amazon.s3_bucket

      if target.to_s.match(/there is no bucket named/i)
        puts "== The S3 Bucket is not setup for this site."
        puts "==== Please run 'bundle exec skypager setup deployment' to begin this process"
        return
      end

      payload = prepare_deploy(build_dir)

      if payload.empty?
        puts "== Nothing to deploy"
      end

      payload.each do |hash|
        body = hash[:node].read

        if target.files.create(key: hash[:key], body: body, public: true)
          rel = hash[:key]
          puts "== Deployed #{ body.length } bytes to #{rel}"
          build_manifest[rel] = hash[:digest]
        end
      end

      set(:build_manifest, build_manifest)
      payload.length
    end

    def use_cdn?
      deploy_options.fetch(:custom_domain).to_s.length > 0 || (get(:use_cdn) == true || get(:use_cdn) == "yes" || get(:use_cdn) == "true")
    end

    def remove_build_manifest
      set(:build_manifest, @build_manifest = {})
    end

    # TODO
    # Let's store build manifests in their own files
    def build_manifest
      @build_manifest ||= get!(:build_manifest) || {}
    end

    def rebuild_manifest(build_dir)
      walker = lambda do |node|
        next if node.to_s == ".DS_Store"

        if node.directory?
          Array(node.children).each(&walker)
        elsif node.file?
          rel = node.relative_path_from(build_dir).to_s
          digest = Digest::MD5.hexdigest(node.read)
          build_manifest[rel] = digest
        end
      end

      build_dir.children.each(&walker)
      set(:build_manifest, build_manifest)
    end

    def paths_to_invalidate(app)
      app.sitemap.resources.map {|r| "/#{r.request_path}" }
    end

    def after_setup
      if use_cdn? && !has_cdn?
        Skypager.amazon.create_cdn_for(self)
      end

      if Skypager.config.dnsimple_setup?
        setup_cname() unless custom_domain?
      else
        puts "You will need to setup the DNS for this bucket yourself."

        if has_cdn?
          puts "You will want to configure the DNS for #{ custom_domain }"
          puts "Create a CNAME and point to #{ cname_value }"
        else
          puts "Create a CNAME for #{ cname_value } on #{ Skypager.config.domain } and point to #{ bucket_host }"
        end
      end
    end

    def dns_domain
      Skypager.dns.domain
    end

    def dns_domain_records
      Skypager.dns.domain_records
    end

    def has_cname_record?
      get!(:has_cname)
    end

    def custom_domain?
      "#{get!(:custom_domain) || deploy_options.custom_domain}".to_s.length > 0
    end

    def deploy_options_config_string
      custom_domain = get!(:custom_domain)

      cfg = []

      if custom_domain
        cfg.push(":custom_domain => '#{ custom_domain }'")
        cfg.push(":use_cdn => true")
      else
        cfg.push(":domain => '#{ Skypager.config.domain }'")
      end

      cfg.push(":bucket_name => '#{ bucket_name }'")

      "\ndeploy_to(:amazon, #{cfg.join(", ")})\n"
    end

    def bucket_key
      fetch(:bucket_key) { bucket && bucket.key }
    end

    def bucket_name
      bucket_key
    end

    def custom_domain
      value = get!(:custom_domain)
      value = deploy_options.custom_domain if value.nil? || value.to_s.length > 0
      value
    end

    def fetch(key, &block)
      @config.fetch(key, &block)
    end

    def bucket_url
      @bucket_url ||= fetch(:bucket_url) { Skypager.amazon.website_url_for(bucket) }
    end

    def bucket_host
      URI.parse(bucket_url).host
    end

    def setup_cname
      Skypager.dns.setup_cname(cname_value, bucket_host)
    end

    def cname_value
      if custom_domain && has_cdn?
        cdn.domain
      else
        name
      end
    end

    def description
      "#{ name }.#{ Skypager.config.domain }"
    end

    def has_cdn?
      @cdn_distribution ||= Skypager.amazon.cdn.distributions.find do |cdn|
        cdn.comment == description
      end
    end

    def cdn
      @cdn ||= Skypager.amazon.cdn.distributions.find do |cdn|
        cdn.comment == description
      end
    end

    def aliases
      if custom_domain
        [custom_domain, "www.#{ custom_domain }"]
      end
    end


  end
end
